#include "common.hpp"
#include "array.hpp"
#include "action.hpp"

using namespace std::complex_literals;
using namespace array;

template <typename Scalar, typename Shape>
void dump(ArrayView<Scalar,Shape> Q)
{
    for (size_t i = 0; i < Q.dim(0); ++i) {
        for (size_t j = 0; j < Q.dim(1); ++j) {
            for (size_t k = 0; k < Q.dim(2); ++k) {
                for (size_t l = 0; l < Q.dim(3); ++l) {
                    std::cout << i << "," << j << "," << k << "," << l << " = " << Q(i,j,k,l) << std::endl;
                }
            }
        }
    }

    for (size_t i = 0; i < Q.size(); ++i) {
        std::cout << Q.data()[i] << std::endl;
    }
}

int main0()
{
    Array<Complex,Shape<2,2,2,2,2>,std::vector<double> > Qval;

    Array<ADComplex,Shape<2,2,2,2,2>,std::vector<ADDouble> > Q;
    Array<Complex,Shape<2,2,4,4,4> > U;

    for (size_t i = 0; i < Qval.dim(0); ++i) {
        for (size_t j = 0; i < Qval.dim(1); ++i) {
            Qval.part(i,j).fill(0.1 + 0.2*i + 0.3i * (double)j);
            Q.part(i,j).fill(0.1 + 0.2*i + 0.3i * (double)j);
        }
    }

    for (size_t i = 0; i < U.dim(0); ++i) {
        for (size_t j = 0; j < U.dim(1); ++j) {
            for (size_t k = 0; k < U.dim(2); ++k) {
                U.submatrix(i,j,k) <<
                    1, 0, 0, 0,
                    0, 1, 0, 0,
                    0, 0, 1, 0,
                    0, 0, 0, 1;
            }
        }
    }

    CppAD::Independent(Q.storage());

    Array<ADComplex,Shape<1>,std::vector<ADDouble> > res;

    ArrayView<const Complex,Shape<2,2,4,4,4> > U_const = U;
    res(0u) = S_2(Q, U_const, 0.1);

    CppAD::ADFun<double> f(Q.storage(), res.storage());

    std::vector<double> x(Q.storage().size());
    for (size_t i = 0; i < Q.storage().size(); ++i)
        x[i] = CppAD::Value(Q.storage()[i]);

    auto jac = f.Jacobian(x);

    for (size_t i = 0; i < jac.size(); ++i)
        std::cout << jac[i] << std::endl;

    return 0;
}

int main()
{
    Array<ADDouble,Shape<8,8> > Q;

    for (size_t i = 0; i < Q.dim(0); ++i) {
        for (size_t j = 0; i < Q.dim(1); ++i) {
            Q(i,j) = 0;
        }
    }

    CppAD::Independent(Q.storage());

    Array<ADDouble,Shape<1> > res;
    res(0u) = laplacian(Q);

    CppAD::ADFun<double> f(Q.storage(), res.storage());

    std::vector<double> x(Q.storage().size());
    for (size_t i = 0; i < Q.storage().size(); ++i)
        x[i] = CppAD::Value(Q.storage()[i]);

    std::cout << "jac =" << std::endl;
    auto jac = f.Jacobian(x);
    for (size_t i = 0; i < jac.size(); ++i)
        std::cout << jac[i] << std::endl;

    std::cout << "hess =" << std::endl;
    auto hess = f.Hessian(x, 0);
    for (size_t i = 0; i < hess.size();) {
        for (size_t j = 0; j < jac.size() && i < hess.size(); ++j, ++i)
            std::cout << hess[i] << " ";
        std::cout << std::endl;
    }

    return 0;
}
